"""Abstract base classes for different parameters."""

import abc
import dataclasses
import functools
import math
import typing

try:
    import quicktions as fractions  # type: ignore
except ImportError:
    import fractions  # type: ignore

from mutwo.parameters import pitches_constants
from mutwo.parameters import volumes_constants
from mutwo.utilities import constants
from mutwo.utilities import tools

__all__ = ("Pitch", "Volume", "PlayingIndicator", "NotationIndicator")


class Parameter(abc.ABC):
    """Abstract base class for any parameter class."""

    pass


ParameterType = object
DurationType = constants.Real


@functools.total_ordering  # type: ignore
class Pitch(Parameter):
    """Abstract base class for any pitch class.

    If the user wants to define a new pitch class, the abstract
    property :attr:`frequency` has to be overridden.
    """

    # conversion methods between different pitch describing units
    @staticmethod
    def hertz_to_cents(frequency0: constants.Real, frequency1: constants.Real) -> float:
        """Calculates the difference in cents between two frequencies.

        :param frequency0: The first frequency in Hertz.
        :param frequency1: The second frequency in Hertz.
        :return: The difference in cents between the first and the second
            frequency.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Pitch.hertz_to_cents(200, 400)
        1200.0
        """
        return float(1200 * math.log(frequency1 / frequency0, 2))

    @staticmethod
    def ratio_to_cents(ratio: fractions.Fraction) -> float:
        """Converts a frequency ratio to its respective cent value.

        :param ratio: The frequency ratio which cent value shall be
            calculated.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Pitch.ratio_to_cents(fractions.Fraction(3, 2))
        701.9550008653874
        """
        return pitches_constants.CENT_CALCULATION_CONSTANT * math.log10(ratio)

    @staticmethod
    def cents_to_ratio(cents: constants.Real) -> fractions.Fraction:
        """Converts a cent value to its respective frequency ratio.

        :param cents: Cents that shall be converted to a frequency ratio.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Pitch.cents_to_ratio(1200)
        Fraction(2, 1)
        """
        return fractions.Fraction(
            10 ** (cents / pitches_constants.CENT_CALCULATION_CONSTANT)
        )

    @staticmethod
    def hertz_to_midi_pitch_number(frequency: constants.Real) -> float:
        """Converts a frequency in hertz to its respective midi pitch.

        :param frequency: The frequency that shall be translated to a midi pitch
            number.
        :return: The midi pitch number (potentially a floating point number if the
            entered frequency isn't on the grid of the equal divided octave tuning
            with a = 440 Hertz).

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Pitch.hertz_to_midi_pitch_number(440)
        69.0
        >>> abc.Pitch.hertz_to_midi_pitch_number(440 * 3 / 2)
        75.98044999134612
        """
        closest_frequency_index = tools.find_closest_index(
            frequency, pitches_constants.MIDI_PITCH_FREQUENCIES
        )
        closest_frequency = pitches_constants.MIDI_PITCH_FREQUENCIES[
            closest_frequency_index
        ]
        closest_midi_pitch_number = pitches_constants.MIDI_PITCH_NUMBERS[
            closest_frequency_index
        ]
        difference_in_cents = Pitch.hertz_to_cents(frequency, closest_frequency)
        return float(closest_midi_pitch_number + (difference_in_cents / 100))

    # properties
    @property
    @abc.abstractmethod
    def frequency(self) -> float:
        """The frequency in Hertz of the pitch."""
        raise NotImplementedError

    @property
    def midi_pitch_number(self) -> float:
        """The midi pitch number (from 0 to 127) of the pitch."""
        return self.hertz_to_midi_pitch_number(self.frequency)

    # comparison methods
    def __lt__(self, other: "Pitch") -> bool:
        return self.frequency < other.frequency

    def __eq__(self, other: object) -> bool:
        try:
            return self.frequency == other.frequency  # type: ignore
        except AttributeError:
            return False


@functools.total_ordering  # type: ignore
class Volume(Parameter):
    """Abstract base class for any volume class.

    If the user wants to define a new volume class, the abstract
    property :attr:`` has to be overridden.
    """

    @staticmethod
    def decibel_to_amplitude_ratio(
        decibel: constants.Real, reference_amplitude: constants.Real = 1
    ) -> float:
        """Convert decibel to amplitude ratio.

        :param decibel: The decibel number that shall be converted.
        :param reference_amplitude: The amplitude for decibel == 0.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Volume.decibel_to_amplitude_ratio(0)
        1
        >>> abc.Volume.decibel_to_amplitude_ratio(-6)
        0.5011872336272722
        >>> abc.Volume.decibel_to_amplitude_ratio(0, reference_amplitude=0.25)
        0.25
        """
        return float(reference_amplitude * (10 ** (decibel / 20)))

    @staticmethod
    def decibel_to_power_ratio(
        decibel: constants.Real, reference_amplitude: constants.Real = 1
    ) -> float:
        """Convert decibel to power ratio.

        :param decibel: The decibel number that shall be converted.
        :param reference_amplitude: The amplitude for decibel == 0.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Volume.decibel_to_power_ratio(0)
        1
        >>> abc.Volume.decibel_to_power_ratio(-6)
        0.251188643150958
        >>> abc.Volume.decibel_to_power_ratio(0, reference_amplitude=0.25)
        0.25
        """
        return float(reference_amplitude * (10 ** (decibel / 10)))

    @staticmethod
    def amplitude_ratio_to_decibel(
        amplitude: constants.Real, reference_amplitude: constants.Real = 1
    ) -> float:
        """Convert amplitude ratio to decibel.

        :param amplitude: The amplitude that shall be converted.
        :param reference_amplitude: The amplitude for decibel == 0.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Volume.amplitude_ratio_to_decibel(1)
        0
        >>> abc.Volume.amplitude_ratio_to_decibel(0)
        inf
        >>> abc.Volume.amplitude_ratio_to_decibel(0.5)
        -6.020599913279624
        """
        if amplitude == 0:
            return float("-inf")
        else:
            return float(20 * math.log10(amplitude / reference_amplitude))

    @staticmethod
    def power_ratio_to_decibel(
        amplitude: constants.Real, reference_amplitude: constants.Real = 1
    ) -> float:
        """Convert power ratio to decibel.

        :param amplitude: The amplitude that shall be converted.
        :param reference_amplitude: The amplitude for decibel == 0.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Volume.power_ratio_to_decibel(1)
        0
        >>> abc.Volume.power_ratio_to_decibel(0)
        inf
        >>> abc.Volume.power_ratio_to_decibel(0.5)
        -3.010299956639812
        """
        if amplitude == 0:
            return float("-inf")
        else:
            return float(10 * math.log10(amplitude / reference_amplitude))

    @staticmethod
    def amplitude_to_midi_velocity(amplitude: constants.Real) -> int:
        """Convert volume (floating point number from 0 to 1) to midi velocity.

        :param amplitude: A value from 0 to 1 that shall be converted to midi
            velocity (0 to 127).
        :return: The midi velocity.

        The method clips values that are higher than 1 / lower than 0.

        **Example:**

        >>> from mutwo.parameters import abc
        >>> abc.Volume.amplitude_to_midi_velocity(1)
        127
        >>> abc.Volume.amplitude_to_midi_velocity(0)
        0
        """
        velocity = int(round(amplitude * volumes_constants.MAXIMUM_VELOCITY))

        # clip velocity
        if velocity < volumes_constants.MINIMUM_VELOCITY:
            velocity = volumes_constants.MINIMUM_VELOCITY
        if velocity > volumes_constants.MAXIMUM_VELOCITY:
            velocity = volumes_constants.MAXIMUM_VELOCITY

        return velocity

    # properties
    @property
    @abc.abstractmethod
    def amplitude(self) -> constants.Real:
        """The amplitude of the Volume (a number from 0 to 1)."""
        raise NotImplementedError

    @property
    def decibel(self) -> constants.Real:
        """The decibel of the volume (from -120 to 0)"""
        return self.amplitude_ratio_to_decibel(self.amplitude)

    @property
    def midi_velocity(self) -> int:
        """The velocity of the volume (from 0 to 1)."""
        return self.amplitude_to_midi_velocity(self.amplitude)

    # comparison methods
    def __lt__(self, other: "Volume") -> bool:
        return self.amplitude < other.amplitude

    def __eq__(self, other: object) -> bool:
        try:
            return self.amplitude == other.amplitude  # type: ignore
        except AttributeError:
            return False


@dataclasses.dataclass()  # type: ignore
class Indicator(Parameter):
    @property
    @abc.abstractmethod
    def is_active(self) -> bool:
        raise NotImplementedError()

    def get_arguments_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            key: getattr(self, key) for key in self.__dataclass_fields__.keys()  # type: ignore
        }


class PlayingIndicator(Indicator):
    pass


class ExplicitPlayingIndicator(PlayingIndicator):
    def __init__(self, is_active: bool = False):
        self.is_active = is_active

    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.is_active)

    def get_arguments_dict(self) -> typing.Dict[str, typing.Any]:
        return {"is_active": self.is_active}

    @property
    def is_active(self) -> bool:
        return self._is_active

    @is_active.setter
    def is_active(self, is_active: bool):
        self._is_active = is_active


@dataclasses.dataclass()
class ImplicitPlayingIndicator(PlayingIndicator):
    @property
    def is_active(self) -> bool:
        return all(
            tuple(
                argument is not None for argument in self.get_arguments_dict().values()
            )
        )


class NotationIndicator(Indicator):
    @property
    def is_active(self) -> bool:
        return all(
            tuple(
                argument is not None for argument in self.get_arguments_dict().values()
            )
        )


T = typing.TypeVar("T", PlayingIndicator, NotationIndicator)


@dataclasses.dataclass(frozen=True)
class IndicatorCollection(typing.Generic[T]):
    def get_all_indicator(self) -> typing.Tuple[T, ...]:
        return tuple(
            getattr(self, key)
            for key in self.__dataclass_fields__.keys()  # type: ignore
        )

    def get_indicator_dict(self) -> typing.Dict[str, Indicator]:
        return {key: getattr(self, key) for key in self.__dataclass_fields__.keys()}  # type: ignore



"""Build tuning files for `Lilypond <https://lilypond.org/>`_ extension `Ekmelily <http://www.ekmelic-music.org/en/extra/ekmelily.htm>`_.

By default the smallest step which Lilypond supports is one quartertone. With
the help of Ekmelily it is easily possible to add more complex micro- or
macrotonal tunings to Lilypond. The converter in this module aims to make it easier
to build tuning files to be used with the 'ekmel-main.ily' script from Thomas Richter.

**Disclaimer:**

For now the converters only support making notation tables for English note names.
"""

import dataclasses
import itertools
import typing
import warnings

try:
    import quicktions as fractions  # type: ignore
except ImportError:
    import fractions  # type: ignore

from mutwo.converters import abc as converters_abc
from mutwo.converters import frontends
from mutwo.parameters import pitches
from mutwo.parameters import pitches_constants
from mutwo.utilities import constants

__all__ = (
    "EkmelilyAccidental",
    "EkmelilyTuningFileConverter",
    "HEJIEkmelilyTuningFileConverter",
)


@dataclasses.dataclass(frozen=True)
class EkmelilyAccidental(object):
    """Representation of an Ekmelily accidental.

    :param accidental_name: The name of the accidental that
        follows after the diatonic pitch name (e.g. 's' or 'qf')
    :type accidental_name: str
    :param accidental_glyphs: The name of accidental glyphs that should
        appear before the notehead. For
        a list of available glyphs, check the documentation of
        `Ekmelos <http://www.ekmelic-music.org/en/extra/ekmelos.htm>`_.
        Furthermore one can find mappings from mutwo data to Ekmelos glyph
        names in :const:`~mutwo.converters.frontends.ekmelily_constants.PRIME_AND_EXPONENT_AND_TRADITIONAL_ACCIDENTAL_TO_ACCIDENTAL_GLYPH`
        and :const:`~mutwo.converters.frontends.ekmelily_constants.TEMPERED_ACCIDENTAL_TO_ACCIDENTAL_GLYPH`.
    :type accidental_glyphs: typing.Tuple[str, ...]
    :param deviation_in_cents: How many cents shall an altered pitch differ from
        its diatonic / natural counterpart.
    :type deviation_in_cents: float
    :param available_diatonic_pitch_indices: Sometimes one
        may want to define accidentals which are only available for certain
        diatonic pitches. For this case, one can use this argument
        and specify all diatonic pitches which should know this
        accidental. If this argument keeps undefined, the accidental
        will be added to all seven diatonic pitches.
    :type available_diatonic_pitch_indices: typing.Optional[typing.Tuple[int, ...]], optional

    **Example:**

    >>> from mutwo.converter.frontends import ekmelily
    >>> natural = ekmelily.EkmelilyAccidental('', ("#xE261",), 0)
    >>> sharp = ekmelily.EkmelilyAccidental('s', ("#xE262",), 100)
    >>> flat = ekmelily.EkmelilyAccidental('f', ("#xE260",), -100)
    """

    accidental_name: str
    accidental_glyphs: typing.Tuple[str, ...]
    deviation_in_cents: float
    available_diatonic_pitch_indices: typing.Optional[typing.Tuple[int, ...]] = None

    def __hash__(self) -> int:
        return hash(
            (
                self.accidental_name,
                self.accidental_glyphs,
                self.deviation_in_cents,
                self.available_diatonic_pitch_indices,
            )
        )


class EkmelilyTuningFileConverter(converters_abc.Converter):
    """Build Ekmelily tuning files from Ekmelily accidentals.

    :param path: Path where the new Ekmelily tuning file shall be written.
        The suffix '.ily' is recommended, but not necessary.
    :type path: str
    :param ekmelily_accidentals: A sequence which contains all
        :class:`EkmelilyAccidental` that shall be written to the tuning file,
    :type ekmelily_accidentals: typing.Sequence[EkmelilyAccidental]
    :param global_scale: From the `Lilypond documentation <https://lilypond.org/doc/v2.20/Documentation/notation/scheme-functions>`_:
        "This determines the tuning of pitches with no accidentals or key signatures.
        The first pitch is c. Alterations are calculated relative to this scale.
        The number of pitches in this scale determines the number of scale steps
        that make up an octave. Usually the 7-note major scale."
    :type global_scale: typing.Tuple[fractions.Fraction, ...], optional

    **Example:**

    >>> from mutwo.converter.frontends import ekmelily
    >>> natural = ekmelily.EkmelilyAccidental('', ("#xE261",), 0)
    >>> sharp = ekmelily.EkmelilyAccidental('s', ("#xE262",), 100)
    >>> flat = ekmelily.EkmelilyAccidental('f', ("#xE260",), -100)
    >>> eigth_tone_sharp = ekmelily.EkmelilyAccidental('es', ("#xE2C7",), 25)
    >>> eigth_tone_flat = ekmelily.EkmelilyAccidental('ef', ("#xE2C2",), -25)
    >>> converter = ekmelily.EkmelilyTuningFileConverter(
    >>>     'ekme-test.ily', (natural, sharp, flat, eigth_tone_sharp, eigth_tone_flat)
    >>> )
    >>> converter.convert()
    """

    def __init__(
        self,
        path: str,
        ekmelily_accidentals: typing.Sequence[EkmelilyAccidental],
        # should have exactly 7 fractions (one for each diatonic pitch)
        global_scale: typing.Optional[typing.Tuple[fractions.Fraction, ...]] = None,
    ):
        if global_scale is None:
            # set to default 12 EDO, a' = 440 Hertz
            global_scale = frontends.ekmelily_constants.DEFAULT_GLOBAL_SCALE

        global_scale = EkmelilyTuningFileConverter._correct_global_scale(global_scale)

        self._path = path
        self._global_scale = global_scale
        self._ekmelily_accidentals = ekmelily_accidentals
        (
            self._accidental_to_alteration_code_mapping,
            self._alteration_code_to_alteration_fraction_mapping,
        ) = EkmelilyTuningFileConverter._make_accidental_to_alteration_code_mapping_and_alteration_code_to_alteration_fraction_mapping(
            self._ekmelily_accidentals
        )

    # ###################################################################### #
    #                          static methods                                #
    # ###################################################################### #

    @staticmethod
    def _correct_global_scale(
        global_scale: typing.Tuple[fractions.Fraction, ...]
    ) -> typing.Tuple[fractions.Fraction, ...]:
        # Lilypond doesn't allow negative values for first item in global scale.
        # Therefore Mutwo makes sure that the first item isn't a negative number.
        corrected_global_scale = list(global_scale)
        if corrected_global_scale[0] != 0:
            message = (
                "Found value '{}' for first scale degree in global scale. Autoset value"
                " to 0 (Lilypond doesn't allow values != 0 for the first scale degree)"
                .format(corrected_global_scale[0])
            )
            warnings.warn(message)
            corrected_global_scale[0] = fractions.Fraction(0, 1)
        return tuple(corrected_global_scale)

    @staticmethod
    def _deviation_in_cents_to_alteration_fraction(
        deviation_in_cents: constants.Real, max_denominator: int = 1000
    ) -> fractions.Fraction:
        # simplify fraction to avoid too complex calculations during Lilyponds midi
        # render (otherwise Lilypond won't be able to render Midi / it will take
        # too long)
        return fractions.Fraction(deviation_in_cents / 200).limit_denominator(
            max_denominator
        )

    @staticmethod
    def _alteration_fraction_to_deviation_in_cents(
        alteration_fraction: fractions.Fraction,
    ) -> float:
        return float(alteration_fraction * 200)

    @staticmethod
    def _accidental_index_to_alteration_code(accidental_index: int) -> str:
        # convert index of accidental to hex code in a format that is
        # readable by Lilypond
        return "#x{}".format(str(hex(accidental_index))[2:].upper())

    @staticmethod
    def _find_and_group_accidentals_by_specific_deviation_in_cents(
        ekmelily_accidentals: typing.Sequence[EkmelilyAccidental],
        deviation_in_cents: float,
    ) -> typing.Tuple[
        typing.Tuple[EkmelilyAccidental, ...], typing.Tuple[EkmelilyAccidental, ...]
    ]:
        positive, negative = [], []
        for accidental in ekmelily_accidentals:
            if accidental.deviation_in_cents == deviation_in_cents:
                positive.append(accidental)
            elif (
                accidental.deviation_in_cents == -deviation_in_cents
                and deviation_in_cents != 0
            ):
                negative.append(accidental)

        return tuple(positive), tuple(negative)

    @staticmethod
    def _group_accidentals_by_deviations_in_cents(
        ekmelily_accidentals: typing.Sequence[EkmelilyAccidental],
    ) -> typing.Tuple[
        typing.Tuple[float, typing.Tuple[typing.Tuple[EkmelilyAccidental, ...], ...]],
        ...,
    ]:
        """Put all accidentals with the same absolute deviation to the same tuple.

        The first element of each tuple is the absolute deviation in cents,
        the second element is tuple with two elements where the first element
        contains all positive accidentals and the second tuple all negative
        accidentals.
        """

        available_deviations_in_cents = sorted(
            set(
                map(
                    lambda accidental: abs(accidental.deviation_in_cents),
                    ekmelily_accidentals,
                )
            )
        )
        accidentals_grouped_by_deviations_in_cents = tuple(
            (
                deviation_in_cents,
                EkmelilyTuningFileConverter._find_and_group_accidentals_by_specific_deviation_in_cents(
                    ekmelily_accidentals, deviation_in_cents
                ),
            )
            for deviation_in_cents in available_deviations_in_cents
        )

        return accidentals_grouped_by_deviations_in_cents

    @staticmethod
    def _process_single_accidental(
        accidental_to_alteration_code_mapping: typing.Dict[EkmelilyAccidental, str],
        alteration_code_to_alteration_fraction_mapping: typing.Dict[
            str, fractions.Fraction
        ],
        accidental: typing.Optional[EkmelilyAccidental],
        accidental_index: int,
        is_positive: bool,
        absolute_deviation_in_cents: float,
    ):
        alteration_code = EkmelilyTuningFileConverter._accidental_index_to_alteration_code(
            accidental_index
        )

        if accidental:
            accidental_to_alteration_code_mapping.update({accidental: alteration_code})

        if is_positive:
            alteration_fraction = EkmelilyTuningFileConverter._deviation_in_cents_to_alteration_fraction(
                absolute_deviation_in_cents
            )
            alteration_code_to_alteration_fraction_mapping.update(
                {alteration_code: alteration_fraction}
            )

    @staticmethod
    def _process_accidental_pair(
        accidental_to_alteration_code_mapping: typing.Dict[EkmelilyAccidental, str],
        alteration_code_to_alteration_fraction_mapping: typing.Dict[
            str, fractions.Fraction
        ],
        positive_accidental: typing.Optional[EkmelilyAccidental],
        positive_alteration_index: int,
        negative_accidental: typing.Optional[EkmelilyAccidental],
        negative_alteration_index: int,
        absolute_deviation_in_cents: float,
    ) -> None:
        # Add accidental data to accidental_to_alteration_code_mapping
        # and alteration_code_to_alteration_fraction_mapping.

        for is_positive, accidental, accidental_index in (
            (True, positive_accidental, positive_alteration_index),
            (False, negative_accidental, negative_alteration_index),
        ):
            EkmelilyTuningFileConverter._process_single_accidental(
                accidental_to_alteration_code_mapping,
                alteration_code_to_alteration_fraction_mapping,
                accidental,
                accidental_index,
                is_positive,
                absolute_deviation_in_cents,
            )

    @staticmethod
    def _get_accidental_from_accidental_iterator(
        accidental_iterator: typing.Iterator,
    ) -> typing.Optional[EkmelilyAccidental]:
        try:
            accidental = next(accidental_iterator)
        except StopIteration:
            accidental = None

        return accidental

    @staticmethod
    def _process_accidental_group(
        accidental_to_alteration_code_mapping: typing.Dict[EkmelilyAccidental, str],
        alteration_code_to_alteration_fraction_mapping: typing.Dict[
            str, fractions.Fraction
        ],
        nth_alteration: int,
        accidental_group: typing.Tuple[typing.Tuple[EkmelilyAccidental, ...], ...],
        absolute_deviation_in_cents: float,
    ) -> int:
        # Define alteration codes and alteration fractions for accidentals
        # in one accidental group and add the calculated data to both mappings.

        positive_accidentals, negative_accidentals = accidental_group
        positive_accidentals_iterator = iter(positive_accidentals)
        negative_accidentals_iterator = iter(negative_accidentals)
        for _ in range(max((len(positive_accidentals), len(negative_accidentals)))):
            positive_alteration_index, negative_alteration_index = (
                nth_alteration,
                nth_alteration + 1,
            )
            positive_accidental = EkmelilyTuningFileConverter._get_accidental_from_accidental_iterator(
                positive_accidentals_iterator
            )
            negative_accidental = EkmelilyTuningFileConverter._get_accidental_from_accidental_iterator(
                negative_accidentals_iterator
            )

            EkmelilyTuningFileConverter._process_accidental_pair(
                accidental_to_alteration_code_mapping,
                alteration_code_to_alteration_fraction_mapping,
                positive_accidental,
                positive_alteration_index,
                negative_accidental,
                negative_alteration_index,
                absolute_deviation_in_cents,
            )
            nth_alteration += 2

        return nth_alteration

    @staticmethod
    def _make_accidental_to_alteration_code_mapping_and_alteration_code_to_alteration_fraction_mapping(
        ekmelily_accidentals: typing.Sequence[EkmelilyAccidental],
    ) -> typing.Tuple[
        typing.Dict[EkmelilyAccidental, str], typing.Dict[str, fractions.Fraction]
    ]:
        accidentals_grouped_by_deviations_in_cents = EkmelilyTuningFileConverter._group_accidentals_by_deviations_in_cents(
            ekmelily_accidentals
        )
        accidental_to_alteration_code_mapping: typing.Dict[EkmelilyAccidental, str] = {}
        alteration_code_to_alteration_fraction_mapping: typing.Dict[
            str, fractions.Fraction
        ] = {}

        nth_alteration = 0
        for (
            absolute_deviation_in_cents,
            accidental_group,
        ) in accidentals_grouped_by_deviations_in_cents:
            nth_alteration = EkmelilyTuningFileConverter._process_accidental_group(
                accidental_to_alteration_code_mapping,
                alteration_code_to_alteration_fraction_mapping,
                nth_alteration,
                accidental_group,
                absolute_deviation_in_cents,
            )

        return (
            accidental_to_alteration_code_mapping,
            alteration_code_to_alteration_fraction_mapping,
        )

    # ###################################################################### #
    #                          private methods                               #
    # ###################################################################### #

    def _make_tuning_table(self) -> str:
        tuning_table_entries = [
            "  (-1 {})".format(" ".join((str(ratio) for ratio in self._global_scale)))
        ]
        for (
            alteration_code,
            alteration_fraction,
        ) in self._alteration_code_to_alteration_fraction_mapping.items():
            alteration = "({} . {})".format(alteration_code, alteration_fraction)
            tuning_table_entries.append(alteration)

        tuning_table = "ekmTuning = #'(\n{})".format("\n  ".join(tuning_table_entries))
        return tuning_table

    def _get_pitch_entry_from_accidental_and_diatonic_pitch(
        self,
        accidental: EkmelilyAccidental,
        nth_diatonic_pitch: int,
        diatonic_pitch: str,
    ) -> typing.Optional[str]:
        is_addable = True
        if accidental.available_diatonic_pitch_indices is not None:
            is_addable = (
                nth_diatonic_pitch in accidental.available_diatonic_pitch_indices
            )
        if is_addable:
            pitch_name = "{}{}".format(diatonic_pitch, accidental.accidental_name)
            alteration_code = self._accidental_to_alteration_code_mapping[accidental]
            pitch_entry = "({} {} . {})".format(
                pitch_name, nth_diatonic_pitch, alteration_code
            )
            return pitch_entry

        return None

    def _make_languages_table(self) -> str:
        language_table_entries = []

        for accidental in self._ekmelily_accidentals:
            for nth_diatonic_pitch, diatonic_pitch in enumerate(
                pitches_constants.ASCENDING_DIATONIC_PITCH_NAMES
            ):
                pitch_entry = self._get_pitch_entry_from_accidental_and_diatonic_pitch(
                    accidental, nth_diatonic_pitch, diatonic_pitch
                )
                if pitch_entry is not None:
                    language_table_entries.append(pitch_entry)

        # for now: only support english language
        languages_table = "ekmLanguages = #'(\n(english . (\n  {})))".format(
            "\n  ".join(language_table_entries)
        )
        return languages_table

    def _make_notations_table(self) -> str:
        notations_table_entries: typing.List[str] = []

        for accidental in self._ekmelily_accidentals:
            alteration_code = self._accidental_to_alteration_code_mapping[accidental]
            accidental_notation = "({} {})".format(
                alteration_code, " ".join(accidental.accidental_glyphs)
            )
            notations_table_entries.append(accidental_notation)

        notations_table = "ekmNotations = #'(\n(default .(\n  {})))".format(
            "\n  ".join(notations_table_entries)
        )
        return notations_table

    # ###################################################################### #
    #                          public api                                    #
    # ###################################################################### #

    def convert(self):
        """Render tuning file to :attr:`path`."""

        ekmelily_tuning_file = (
            self._make_tuning_table(),
            self._make_languages_table(),
            self._make_notations_table(),
            r'\include "ekmel-main.ily"',
        )

        ekmelily_tuning_file = "\n\n".join(ekmelily_tuning_file)

        with open(self._path, "w") as f:
            f.write(ekmelily_tuning_file)


class HEJIEkmelilyTuningFileConverter(EkmelilyTuningFileConverter):
    """Build Ekmelily tuning files for `Helmholtz-Ellis JI Pitch Notation <https://marsbat.space/pdfs/notation.pdf>`_.

    :param path: Path where the new Ekmelily tuning file shall be written.
        The suffix '.ily' is recommended, but not necessary.
    :type path: str
    :param prime_to_highest_allowed_exponent: Mapping of prime number to
        highest exponent that should occur. Take care not to add
        higher exponents than the HEJI Notation supports. See
        :const:`~mutwo.converters.frontends.ekmelily_constants.DEFAULT_PRIME_TO_HIGHEST_ALLOWED_EXPONENT`
        for the default mapping.
    :type prime_to_highest_allowed_exponent: typing.Dict[int, int], optional
    :param reference_pitch: The reference pitch (1/1). Should be a diatonic
        pitch name (see
        :const:`~mutwo.parameters.pitches_constants.ASCENDING_DIATONIC_PITCH_NAMES`)
        in English nomenclature. For any other reference pitch than 'c', Lilyponds
        midi rendering for pitches with the diatonic pitch 'c' will be slightly
        out of tune (because the first value of `global_scale`
        always have to be 0).
    :type reference_pitch: str, optional
    :param prime_to_heji_accidental_name: Mapping of a prime number
        to a string which indicates the respective prime number in the resulting
        accidental name. See
        :const:`~mutwo.converters.frontends.ekmelily_constants.DEFAULT_PRIME_TO_HEJI_ACCIDENTAL_NAME`
        for the default mapping.
    :type prime_to_heji_accidental_name: typing.Dict[int, str], optional
    :param otonality_indicator: String which indicates that the
        respective prime alteration is otonal. See
        :const:`~mutwo.converters.frontends.ekmelily_constants.DEFAULT_OTONALITY_INDICATOR`
        for the default value.
    :type otonality_indicator: str, optional
    :param utonality_indicator: String which indicates that the
        respective prime alteration is utonal. See
        :const:`~mutwo.converters.frontends.ekmelily_constants.DEFAULT_OTONALITY_INDICATOR`
        for the default value.
    :type utonality_indicator: str, optional
    :param exponent_to_exponent_indicator: Function to convert the
        exponent of a prime number to string which indicates the respective
        exponent. See
        :func:`~mutwo.converters.frontends.ekmelily_constants.DEFAULT_EXPONENT_TO_EXPONENT_INDICATOR`
        for the default function.
    :type exponent_to_exponent_indicator: typing.Callable[[int], str], optional
    :param tempered_pitch_indicator: String which indicates that the
        respective accidental is tempered (12 EDO). See
        :const:`~mutwo.converters.frontends.ekmelily_constants.DEFAULT_TEMPERED_PITCH_INDICATOR`
        for the default value.
    :type tempered_pitch_indicator: str, optional
    """

    def __init__(
        self,
        path: str = None,
        prime_to_highest_allowed_exponent: typing.Optional[
            typing.Dict[int, int]
        ] = None,
        reference_pitch: str = "c",
        prime_to_heji_accidental_name: typing.Optional[typing.Dict[int, str]] = None,
        otonality_indicator: str = None,
        utonality_indicator: str = None,
        exponent_to_exponent_indicator: typing.Callable[[int], str] = None,
        tempered_pitch_indicator: str = None,
    ):
        # set default values
        if path is None:
            path = "ekme-heji-ref-{}.ily".format(reference_pitch)

        if prime_to_highest_allowed_exponent is None:
            prime_to_highest_allowed_exponent = (
                frontends.ekmelily_constants.DEFAULT_PRIME_TO_HIGHEST_ALLOWED_EXPONENT
            )

        if prime_to_heji_accidental_name is None:
            prime_to_heji_accidental_name = (
                frontends.ekmelily_constants.DEFAULT_PRIME_TO_HEJI_ACCIDENTAL_NAME
            )

        if otonality_indicator is None:
            otonality_indicator = (
                frontends.ekmelily_constants.DEFAULT_OTONALITY_INDICATOR
            )

        if utonality_indicator is None:
            utonality_indicator = (
                frontends.ekmelily_constants.DEFAULT_UTONALITY_INDICATOR
            )

        if exponent_to_exponent_indicator is None:
            exponent_to_exponent_indicator = (
                frontends.ekmelily_constants.DEFAULT_EXPONENT_TO_EXPONENT_INDICATOR
            )

        if tempered_pitch_indicator is None:
            tempered_pitch_indicator = (
                frontends.ekmelily_constants.DEFAULT_TEMPERED_PITCH_INDICATOR
            )

        difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches = HEJIEkmelilyTuningFileConverter._find_difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches(
            reference_pitch
        )
        global_scale = HEJIEkmelilyTuningFileConverter._make_global_scale(
            difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches
        )
        ekmelily_accidentals = HEJIEkmelilyTuningFileConverter._make_ekmelily_accidentals(
            difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches,
            prime_to_highest_allowed_exponent,
            prime_to_heji_accidental_name,
            otonality_indicator,
            utonality_indicator,
            exponent_to_exponent_indicator,
            tempered_pitch_indicator,
        )
        super().__init__(path, ekmelily_accidentals, global_scale=global_scale)

    # ###################################################################### #
    #                          static methods                                #
    # ###################################################################### #

    @staticmethod
    def _find_difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches(
        reference_pitch: str,
    ) -> typing.Tuple[float, ...]:
        difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches = []
        reference_pitch_index = pitches_constants.DIATONIC_PITCH_NAME_CYCLE_OF_FIFTHS.index(
            reference_pitch
        )
        for diatonic_pitch_name in pitches_constants.ASCENDING_DIATONIC_PITCH_NAMES:
            pitch_index = pitches_constants.DIATONIC_PITCH_NAME_CYCLE_OF_FIFTHS.index(
                diatonic_pitch_name
            )
            n_exponents_difference_from_reference = pitch_index - reference_pitch_index
            difference_from_tempered_diatonic_pitch = (
                frontends.ekmelily_constants.DIFFERENCE_BETWEEN_PYTHAGOREAN_AND_TEMPERED_FIFTH
                * n_exponents_difference_from_reference
            )
            difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches.append(
                difference_from_tempered_diatonic_pitch
            )

        return tuple(difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches)

    @staticmethod
    def _make_global_scale(
        difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches: typing.Tuple[
            float, ...
        ],
    ) -> typing.Tuple[fractions.Fraction, ...]:
        new_global_scale = []

        for (
            nth_diatonic_pitch,
            difference_in_cents_from_tempered_pitch_class,
        ) in enumerate(
            difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches
        ):
            default_cents_for_diatonic_pitch = EkmelilyTuningFileConverter._alteration_fraction_to_deviation_in_cents(
                frontends.ekmelily_constants.DEFAULT_GLOBAL_SCALE[nth_diatonic_pitch]
            )
            n_cents = (
                default_cents_for_diatonic_pitch
                + difference_in_cents_from_tempered_pitch_class
            )
            alteration_fraction = EkmelilyTuningFileConverter._deviation_in_cents_to_alteration_fraction(
                n_cents, max_denominator=1000
            )
            new_global_scale.append(alteration_fraction)

        return tuple(new_global_scale)

    @staticmethod
    def _make_pythagorean_accidentals() -> typing.Tuple[EkmelilyAccidental, ...]:
        accidentals = []
        for (
            alteration_name
        ) in (
            frontends.ekmelily_constants.PYTHAGOREAN_ACCIDENTAL_TO_CENT_DEVIATIONS.keys()
        ):
            glyph = frontends.ekmelily_constants.PRIME_AND_EXPONENT_AND_TRADITIONAL_ACCIDENTAL_TO_ACCIDENTAL_GLYPH[
                None, None, alteration_name
            ]
            cents_deviation = frontends.ekmelily_constants.PYTHAGOREAN_ACCIDENTAL_TO_CENT_DEVIATIONS[
                alteration_name
            ]
            accidental = EkmelilyAccidental(
                alteration_name, (glyph,), cents_deviation
            )
            accidentals.append(accidental)

        return tuple(accidentals)

    @staticmethod
    def _process_prime(
        prime_to_heji_accidental_name: typing.Dict[int, str],
        pythagorean_accidental: str,
        prime: int,
        exponent: int,
        otonality_indicator: str,
        utonality_indicator: str,
        exponent_to_exponent_indicator: typing.Callable[[int], str],
    ) -> typing.Tuple[str, str, float]:
        glyph_key = (
            (prime, exponent, pythagorean_accidental)
            if prime == 5
            else (prime, exponent, None)
        )
        glyph = frontends.ekmelily_constants.PRIME_AND_EXPONENT_AND_TRADITIONAL_ACCIDENTAL_TO_ACCIDENTAL_GLYPH[
            glyph_key
        ]

        accidental_name = "{}{}{}".format(
            (otonality_indicator, utonality_indicator)[exponent < 0],
            prime_to_heji_accidental_name[prime],
            exponent_to_exponent_indicator(abs(exponent) - 1),
        )

        cents_deviation = pitches.JustIntonationPitch(
            pitches_constants.DEFAULT_PRIME_TO_COMMA[prime].ratio ** exponent
        ).cents
        return accidental_name, glyph, cents_deviation

    @staticmethod
    def _make_higher_prime_accidental(
        pythagorean_accidental: str,
        pythagorean_accidental_cents_deviation: float,
        exponents: typing.Tuple[int, ...],
        prime_to_highest_allowed_exponent: typing.Dict[int, int],
        prime_to_heji_accidental_name: typing.Dict[int, str],
        otonality_indicator: str,
        utonality_indicator: str,
        exponent_to_exponent_indicator: typing.Callable[[int], str],
    ) -> EkmelilyAccidental:
        accidental_parts = ["{}".format(pythagorean_accidental)]
        cents_deviation = float(pythagorean_accidental_cents_deviation)
        glyphs = []
        for prime, exponent in zip(
            prime_to_highest_allowed_exponent.keys(), exponents,
        ):
            if exponent != 0:
                (
                    accidental_change,
                    glyph,
                    cents_deviation_change,
                ) = HEJIEkmelilyTuningFileConverter._process_prime(
                    prime_to_heji_accidental_name,
                    pythagorean_accidental,
                    prime,
                    exponent,
                    otonality_indicator,
                    utonality_indicator,
                    exponent_to_exponent_indicator,
                )
                accidental_parts.append(accidental_change)
                glyphs.append(glyph)
                cents_deviation += cents_deviation_change

        # add traditional accidentals (sharp, flat, etc.) if no syntonic
        # comma is available (if there is any syntonic comma there is
        # already the necessary pythagorean accidental)
        if exponents[0] == 0:
            glyphs.insert(
                0,
                frontends.ekmelily_constants.PRIME_AND_EXPONENT_AND_TRADITIONAL_ACCIDENTAL_TO_ACCIDENTAL_GLYPH[
                    (None, None, pythagorean_accidental)
                ],
            )

        # start with highest primes
        glyphs.reverse()

        accidental_name = "".join(accidental_parts)

        new_accidental = EkmelilyAccidental(
            accidental_name, tuple(glyphs), round(cents_deviation, 4)
        )
        return new_accidental

    @staticmethod
    def _make_accidentals_for_higher_primes(
        prime_to_highest_allowed_exponent: typing.Dict[int, int],
        prime_to_heji_accidental_name: typing.Dict[int, str],
        otonality_indicator: str,
        utonality_indicator: str,
        exponent_to_exponent_indicator: typing.Callable[[int], str],
    ) -> typing.Tuple[EkmelilyAccidental, ...]:
        allowed_exponents = tuple(
            tuple(range(-maxima_exponent, maxima_exponent + 1))
            for _, maxima_exponent in sorted(
                prime_to_highest_allowed_exponent.items(),
                key=lambda prime_to_maxima_exponent: prime_to_maxima_exponent[0],
            )
        )
        accidentals = []

        for exponents in itertools.product(*allowed_exponents):
            if any(tuple(exp != 0 for exp in exponents)):
                for (
                    pythagorean_accidental,
                    pythagorean_accidental_cents_deviation,
                ) in (
                    frontends.ekmelily_constants.PYTHAGOREAN_ACCIDENTAL_TO_CENT_DEVIATIONS.items()
                ):
                    accidental = HEJIEkmelilyTuningFileConverter._make_higher_prime_accidental(
                        pythagorean_accidental,
                        pythagorean_accidental_cents_deviation,
                        exponents,
                        prime_to_highest_allowed_exponent,
                        prime_to_heji_accidental_name,
                        otonality_indicator,
                        utonality_indicator,
                        exponent_to_exponent_indicator,
                    )
                    accidentals.append(accidental)

        return tuple(accidentals)

    @staticmethod
    def _make_tempered_accidentals(
        difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches: typing.Tuple[
            float, ...
        ],
        tempered_pitch_indicator: str,
    ) -> typing.Tuple[EkmelilyAccidental, ...]:
        accidentals = []

        for (
            nth_diatonic_pitch,
            difference_in_cents_from_tempered_pitch_class,
        ) in enumerate(
            difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches,
        ):
            for (
                tempered_accidental,
                accidental_glyph,
            ) in (
                frontends.ekmelily_constants.TEMPERED_ACCIDENTAL_TO_ACCIDENTAL_GLYPH.items()
            ):

                accidental_name = "{}{}".format(
                    tempered_accidental, tempered_pitch_indicator
                )
                deviation_in_cents = (
                    frontends.ekmelily_constants.TEMPERED_ACCIDENTAL_TO_CENT_DEVIATION[
                        tempered_accidental
                    ]
                    - difference_in_cents_from_tempered_pitch_class
                )

                accidental = EkmelilyAccidental(
                    accidental_name,
                    (accidental_glyph,),
                    deviation_in_cents,
                    available_diatonic_pitch_indices=(nth_diatonic_pitch,),
                )
                accidentals.append(accidental)

        return tuple(accidentals)

    @staticmethod
    def _make_ekmelily_accidentals(
        difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches: typing.Tuple[
            float, ...
        ],
        prime_to_highest_allowed_exponent: typing.Dict[int, int],
        prime_to_heji_accidental_name: typing.Dict[int, str],
        otonality_indicator: str,
        utonality_indicator: str,
        exponent_to_exponent_indicator: typing.Callable[[int], str],
        tempered_pitch_indicator: str,
    ) -> typing.Tuple[EkmelilyAccidental, ...]:

        pythagorean_accidentals = (
            HEJIEkmelilyTuningFileConverter._make_pythagorean_accidentals()
        )

        accidentals_for_higher_primes = HEJIEkmelilyTuningFileConverter._make_accidentals_for_higher_primes(
            prime_to_highest_allowed_exponent,
            prime_to_heji_accidental_name,
            otonality_indicator,
            utonality_indicator,
            exponent_to_exponent_indicator,
        )

        tempered_accidentals = HEJIEkmelilyTuningFileConverter._make_tempered_accidentals(
            difference_in_cents_from_tempered_pitch_class_for_diatonic_pitches,
            tempered_pitch_indicator,
        )

        return (
            pythagorean_accidentals
            + accidentals_for_higher_primes
            + tempered_accidentals
        )



import numbers

from mutwo import events
from mutwo import parameters


class ValueEvent(events.basic.SimpleEvent):
    def __init__(
        self,
        duration: parameters.abc.DurationType,
        value: numbers.Number,
    ):
        self.value = value
        super().__init__(duration)



"""Render sound files from mutwo data via Csound.

Csound is a `"domain-specific computer programming language
for audio programming" <http://www.csounds.com/>`_.
"""

import numbers
import os
import typing
import warnings

import natsort  # type: ignore

from mutwo import converters
from mutwo import events
from mutwo import parameters
from mutwo.utilities import constants

__all__ = ("CsoundScoreConverter", "CsoundConverter")

SupportedPFieldTypes = typing.Union[constants.Real, str]
SupportedPFieldTypesForTypeChecker = typing.Union[numbers.Real, str]
PFieldFunction = typing.Callable[[events.basic.SimpleEvent], SupportedPFieldTypes]
PFieldDict = typing.Dict[str, typing.Optional[PFieldFunction]]


class CsoundScoreConverter(converters.abc.EventConverter):
    """Class to convert mutwo events to a Csound score file.

    :param path: where to write the csound score file
    :param pfield: p-field / p-field-extraction-function pairs.

    This class helps generating score files for the `"domain-specific computer
    programming language for audio programming" Csound <http://www.csounds.com/>`_.

    :class:`CsoundScoreConverter` extracts data from mutwo Events and assign it to
    specific p-fields. The mapping of Event attributes to p-field values has
    to be defined by the user via keyword arguments during class initialization.

    By default, mutwo already maps the following p-fields to the following values:

    - p1 (instrument name) to 1
    - p2 (start time) to the absolute start time of the event
    - p3 (duration) to the :attr:`duration` attribute of the event

    If p2 shall be assigned to the absolute entry delay of the event,
    it has to be set to None.

    The :class:`CsoundScoreConverter` ignores any p-field that returns
    any unsupported p-field type (anything else than a string
    or a number). If the returned type is a string, :class:`CsoundScoreConverter`
    automatically adds quotations marks around the string in the score file.

    All p-fields can be overwritten in the following manner:

    >>> my_converter = CsoundScoreConverter(
    >>>     path="my_csound_score.sco",
    >>>     p1=lambda event: 2,
    >>>     p4=lambda event: event.pitch.frequency,
    >>>     p5=lambda event: event.volume
    >>> )

    For easier debugging of faulty score files, :mod:`mutwo` adds annotations
    when a new :class:`SequentialEvent` or a new :class:`SimultaneousEvent`
    starts.
    """

    _default_p_fields: PFieldDict = {
        "p1": lambda event: 1,  # default instrument name "1"
        "p2": None,  # default to absolute start time
        "p3": lambda event: event.duration  # type: ignore
        if event.duration > 0
        else None,  # default key for duration
    }

    def __init__(self, path: str, **pfield: PFieldFunction):
        concatenated_p_fields: PFieldDict = dict([])
        for default_p_field, default_p_field_function in self._default_p_fields.items():
            if default_p_field not in pfield:
                concatenated_p_fields.update(
                    {default_p_field: default_p_field_function}
                )

        concatenated_p_fields.update(pfield)
        self.pfields = self._generate_pfield_mapping(concatenated_p_fields)
        self.path = path

    # ###################################################################### #
    #                          static methods                                #
    # ###################################################################### #

    @staticmethod
    def _generate_pfield_mapping(
        pfield_key_to_function_mapping: PFieldDict,
    ) -> typing.Tuple[typing.Optional[PFieldFunction], ...]:
        """Maps p-fields to their respective p_field_function."""

        sorted_pfield_keys = natsort.natsorted(pfield_key_to_function_mapping.keys())
        pfields = []
        for key0, key1 in zip(sorted_pfield_keys, sorted_pfield_keys[1:]):
            number0, number1 = (int(pfield_name[1:]) for pfield_name in (key0, key1))
            try:
                assert number0 >= 0
            except AssertionError:
                message = (
                    "Can't assign p-field '{}'. P-field number has to bigger than 0."
                    .format(key0)
                )
                raise ValueError(message)

            pfields.append(pfield_key_to_function_mapping[key0])

            difference = number1 - number0
            if difference > 1:
                for _ in range(difference - 1):
                    pfields.append(lambda event: 0)
                message = (
                    "Couldn't find any mapping for p-fields between '{}' and '{}'. "
                    .format(key0, key1)
                )
                message += "Assigned these p-fields to 0."
                warnings.warn(message)

        pfields.append(pfield_key_to_function_mapping[key1])
        return tuple(pfields)

    @staticmethod
    def _process_p_field_value(
        nth_p_field: int, p_field_value: typing.Any
    ) -> typing.Optional[str]:
        """Makes sure pfield value is of correct type & adds quotation marks for str."""

        if isinstance(p_field_value, SupportedPFieldTypesForTypeChecker.__args__):  # type: ignore
            # silently adding quotation marks
            if isinstance(p_field_value, str):
                p_field_value = '"{}"'.format(p_field_value)

            else:
                p_field_value = "{}".format(p_field_value)

            return p_field_value

        else:
            ignored_p_field = nth_p_field + 1
            message = (
                "Can't assign returned value '{}' of type '{}' to p-field {}.".format(
                    p_field_value, type(p_field_value), ignored_p_field
                )
            )
            message += " Supported types for p-fields include '{}'. ".format(
                repr(SupportedPFieldTypes)
            )
            message += "Ignored p-field {}.".format(ignored_p_field)
            warnings.warn(message)
            return None

    # ###################################################################### #
    #           private methods (conversion of different event types)        #
    # ###################################################################### #

    def _convert_simple_event(
        self,
        simple_event: events.basic.SimpleEvent,
        absolute_entry_delay: parameters.abc.DurationType,
    ) -> typing.Tuple[str, ...]:
        """Extract p-field data from simple event and write one Csound-Score line."""

        csound_score_line = "i"
        for nth_p_field, p_field_function in enumerate(self.pfields):
            # special case of absolute start time initialization
            if nth_p_field == 1 and p_field_function is None:
                csound_score_line += " {}".format(absolute_entry_delay)

            else:
                try:
                    p_field_value = p_field_function(simple_event)  # type: ignore

                except AttributeError:
                    # if attribute couldn't be found, just make a rest
                    return tuple([])

                p_field_value = CsoundScoreConverter._process_p_field_value(
                    nth_p_field, p_field_value
                )
                if p_field_value is not None:
                    csound_score_line += " {}".format(p_field_value)

        return (csound_score_line,)

    def _convert_sequential_event(
        self,
        sequential_event: events.basic.SequentialEvent,
        absolute_entry_delay: parameters.abc.DurationType,
    ) -> typing.Tuple[str, ...]:
        csound_score_lines = [
            converters.frontends.csound_constants.SEQUENTIAL_EVENT_ANNOTATION
        ]
        csound_score_lines.extend(
            super()._convert_sequential_event(sequential_event, absolute_entry_delay)
        )

        for _ in range(
            converters.frontends.csound_constants.N_EMPTY_LINES_AFTER_COMPLEX_EVENT
        ):
            csound_score_lines.append("")

        return tuple(csound_score_lines)

    def _convert_simultaneous_event(
        self,
        simultaneous_event: events.basic.SimultaneousEvent,
        absolute_entry_delay: parameters.abc.DurationType,
    ) -> typing.Tuple[str, ...]:
        csound_score_lines = [
            converters.frontends.csound_constants.SIMULTANEOUS_EVENT_ANNOTATION
        ]
        csound_score_lines.extend(
            super()._convert_simultaneous_event(
                simultaneous_event, absolute_entry_delay
            )
        )
        for _ in range(
            converters.frontends.csound_constants.N_EMPTY_LINES_AFTER_COMPLEX_EVENT
        ):
            csound_score_lines.append("")
        return tuple(csound_score_lines)

    # ###################################################################### #
    #                             public api                                 #
    # ###################################################################### #

    def convert(self, event_to_convert: events.abc.Event) -> None:
        """Render csound score file (.sco) from the passed event.

        :param event_to_convert: The event that shall be rendered to a csound score
            file.

        >>> import random
        >>> from mutwo.parameters import pitches
        >>> from mutwo.events import basic
        >>> from mutwo.converters.frontends import csound
        >>> converter = csound.CsoundScoreConverter(
        >>>    path="score.sco", p4=lambda event: event.pitch.frequency
        >>> )
        >>> events = basic.SequentialEvent(
        >>>    [
        >>>        basic.SimpleEvent(random.uniform(0.3, 1.2)) for _ in range(15)
        >>>    ]
        >>> )
        >>> for event in events:
        >>>     event.pitch = pitches.DirectPitch(random.uniform(100, 500))
        >>> converter.convert(events)
        """

        csound_score_lines = self._convert_event(event_to_convert, 0)
        # convert events to strings (where each string represents one csound score line)
        # write csound score lines to file
        with open(self.path, "w") as f:
            f.write("\n".join(csound_score_lines))


class CsoundConverter(converters.abc.Converter):
    """Generate audio files with `Csound <http://www.csounds.com/>`_.

    :param path: where to write the sound file
    :param csound_orchestra_path: Path to the csound orchestra (.orc) file.
    :param csound_score_converter: The :class:`CsoundScoreConverter` that shall be used
        to render the csound score file (.sco) from a mutwo event.
    :param *flag: Flag that shall be added when calling csound. Several of the supported
        csound flags can be found in :mod:`mutwo.converters.frontends.csound_constants`.
    :param remove_score_file: Set to True if :class:`CsoundConverter` shall remove the
        csound score file after rendering. Defaults to False.

    **Disclaimer:** Before using the :class:`CsoundConverter`, make sure
    `Csound <http://www.csounds.com/>`_ has been correctly installed on
    your system.
    """

    def __init__(
        self,
        path: str,
        csound_orchestra_path: str,
        csound_score_converter: CsoundScoreConverter,
        *flag: str,
        remove_score_file: bool = False
    ):
        self.flags = flag
        self.path = path
        self.csound_orchestra_path = csound_orchestra_path
        self.csound_score_converter = csound_score_converter
        self.remove_score_file = remove_score_file

    def convert(self, event_to_convert: events.abc.Event) -> None:
        """Render sound file from the mutwo event.

        :param event_to_convert: The event that shall be rendered.
        """

        self.csound_score_converter.convert(event_to_convert)
        command = "csound -o {}".format(self.path)
        for flag in self.flags:
            command += " {} ".format(flag)
        command += " {} {}".format(
            self.csound_orchestra_path, self.csound_score_converter.path
        )

        os.system(command)

        if self.remove_score_file:
            os.remove(self.csound_score_converter.path)



"""Apply :class:`~mutwo.parameters.abc.PlayingIndicator` on :class:`~mutwo.events.abc.Event`.

"""

import abc
import copy
import typing

from mutwo import converters
from mutwo import events
from mutwo import parameters


__all__ = (
    "PlayingIndicatorConverter",
    "ArpeggioConverter",
    "PlayingIndicatorsConverter",
)


class PlayingIndicatorConverter(converters.abc.Converter):
    """Abstract base class to apply :class:`~mutwo.parameters.abc.PlayingIndicator` on a :class:`~mutwo.events.basic.SimpleEvent`.

    :param simple_event_to_playing_indicators: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a
        :class:`mutwo.parameters.playing_indicators.PlayingIndicatorCollection`
        object. By default it asks the Event for its
        :attr:`~mutwo.events.music.NoteLike.playing_indicators`
        attribute (because by default :class:`mutwo.events.music.NoteLike`
        objects are expected).
        When using different Event classes than :class:`~mutwo.events.music.NoteLike`
        with a different name for their playing_indicators property, this argument
        should be overridden. If the
        function call raises an :obj:`AttributeError` (e.g. if no playing indicator
        collection can be extracted), mutwo will build a playing indicator collection
        from :const:`~mutwo.events.music_constants.DEFAULT_PLAYING_INDICATORS_COLLECTION_CLASS`.
    :type simple_event_to_playing_indicators: typing.Callable[[events.basic.SimpleEvent], parameters.playing_indicators.PlayingIndicatorCollection], optional
    """

    def __init__(
        self,
        simple_event_to_playing_indicators: typing.Callable[
            [events.basic.SimpleEvent],
            parameters.playing_indicators.PlayingIndicatorCollection,
        ] = lambda simple_event: simple_event.playing_indicators,  # type: ignore
    ):
        self._simple_event_to_playing_indicators = simple_event_to_playing_indicators

    @abc.abstractmethod
    def _apply_playing_indicator(
        self,
        simple_event_to_convert: events.basic.SimpleEvent,
        playing_indicators: parameters.playing_indicators.PlayingIndicatorCollection,
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        raise NotImplementedError()

    def convert(
        self, simple_event_to_convert: events.basic.SimpleEvent
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        """Apply PlayingIndicator on simple_event.

        :param simple_event_to_convert: The event which shall be converted.
        :type simple_event_to_convert: events.basic.SimpleEvent
        """

        try:
            playing_indicators = self._simple_event_to_playing_indicators(
                simple_event_to_convert
            )
        except AttributeError:
            playing_indicators = (
                events.music_constants.DEFAULT_PLAYING_INDICATORS_COLLECTION_CLASS()
            )

        return self._apply_playing_indicator(
            simple_event_to_convert, playing_indicators
        )


class ArpeggioConverter(PlayingIndicatorConverter):
    """Apply arpeggio on :class:`~mutwo.events.basic.SimpleEvent`.

    :param duration_for_each_attack: Set how long each attack of the
        Arpeggio lasts. Default to 0.1.
    :type duration_for_each_attack: parameters.abc.DurationType
    :param simple_event_to_pitches: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a tuple that contains pitch objects
        (objects that inherit from :class:`mutwo.parameters.abc.Pitch`).
        By default it asks the Event for its
        :attr:`~mutwo.events.music.NoteLike.pitch_or_pitches` attribute
        (because by default :class:`mutwo.events.music.NoteLike` objects are expected).
        When using different Event classes than :class:`~mutwo.events.music.NoteLike`
        with a different name for their pitch property, this argument
        should be overridden.
        If the function call raises an :obj:`AttributeError` (e.g. if no pitch can be
        extracted), mutwo will assume an event without any pitches.
    :type simple_event_to_pitches: typing.Callable[[events.basic.SimpleEvent], parameters.abc.Pitch], optional
    :param simple_event_to_playing_indicators: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a
        :class:`mutwo.parameters.playing_indicators.PlayingIndicatorCollection`
        object. By default it asks the Event for its
        :attr:`~mutwo.events.music.NoteLike.playing_indicators`
        attribute (because by default :class:`mutwo.events.music.NoteLike`
        objects are expected).
        When using different Event classes than :class:`~mutwo.events.music.NoteLike`
        with a different name for their playing_indicators property, this argument
        should be overridden. If the
        function call raises an :obj:`AttributeError` (e.g. if no playing indicator
        collection can be extracted), mutwo will build a playing indicator collection
        from :const:`~mutwo.events.music_constants.DEFAULT_PLAYING_INDICATORS_COLLECTION_CLASS`.
    :type simple_event_to_playing_indicators: typing.Callable[[events.basic.SimpleEvent], parameters.playing_indicators.PlayingIndicatorCollection,], optional
    :param set_pitches_for_simple_event: Function which assigns
        a list of :class:`~mutwo.parameters.abc.Pitch` objects to a
        :class:`~mutwo.events.basic.SimpleEvent`. By default the
        function assigns the passed pitches to the
        :attr:`~mutwo.events.music.NoteLike.pitch_or_pitches` attribute
        (because by default :class:`mutwo.events.music.NoteLike` objects
        are expected).
    :type set_pitches_for_simple_event: typing.Callable[[events.basic.SimpleEvent, typing.List[parameters.abc.Pitch]], None]
    """

    def __init__(
        self,
        duration_for_each_attack: parameters.abc.DurationType = 0.1,
        simple_event_to_pitches: typing.Callable[
            [events.basic.SimpleEvent], typing.List[parameters.abc.Pitch]
        ] = lambda simple_event: simple_event.pitch_or_pitches,  # type: ignore
        simple_event_to_playing_indicators: typing.Callable[
            [events.basic.SimpleEvent],
            parameters.playing_indicators.PlayingIndicatorCollection,
        ] = lambda simple_event: simple_event.playing_indicators,  # type: ignore
        set_pitches_for_simple_event: typing.Callable[
            [events.basic.SimpleEvent, typing.List[parameters.abc.Pitch]], None
        ] = lambda simple_event, pitch_or_pitches: simple_event.set_parameter(
            "pitch_or_pitches", pitch_or_pitches, set_unassigned_parameter=True
        ),
    ):
        super().__init__(
            simple_event_to_playing_indicators=simple_event_to_playing_indicators
        )
        self._duration_for_each_attack = duration_for_each_attack
        self._simple_event_to_pitches = simple_event_to_pitches
        self._set_pitches_for_simple_event = set_pitches_for_simple_event

    def _apply_arpeggio(
        self,
        simple_event_to_convert: events.basic.SimpleEvent,
        arpeggio: parameters.playing_indicators.Arpeggio,
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        try:
            pitch_or_pitches = list(
                self._simple_event_to_pitches(simple_event_to_convert)
            )
        except AttributeError:
            pitch_or_pitches = []

        # sort pitches according to Arpeggio direction
        pitch_or_pitches.sort(reverse=arpeggio.direction != "up")

        converted_event = events.basic.SequentialEvent(
            [copy.copy(simple_event_to_convert) for _ in pitch_or_pitches]
        )

        # apply pitches on events
        for nth_event, pitch in enumerate(pitch_or_pitches):
            self._set_pitches_for_simple_event(converted_event[nth_event], [pitch])

        # set correct duration for each event
        n_events = len(converted_event)
        duration_of_each_attack = self._duration_for_each_attack
        if n_events * duration_of_each_attack > simple_event_to_convert.duration:
            duration_of_each_attack = simple_event_to_convert.duration / n_events

        for nth_event in range(n_events - 1):
            converted_event[nth_event].duration = duration_of_each_attack

        converted_event[-1].duration -= (
            converted_event.duration - simple_event_to_convert.duration
        )

        return converted_event

    def _apply_playing_indicator(
        self,
        simple_event_to_convert: events.basic.SimpleEvent,
        playing_indicators: parameters.playing_indicators.PlayingIndicatorCollection,
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        try:
            arpeggio = playing_indicators.arpeggio
        except AttributeError:
            arpeggio = parameters.playing_indicators.Arpeggio()

        if arpeggio.is_active:
            return self._apply_arpeggio(simple_event_to_convert, arpeggio)
        else:
            return events.basic.SequentialEvent([copy.copy(simple_event_to_convert)])


class PlayingIndicatorsConverter(converters.abc.EventConverter):
    """Apply :class:`~mutwo.parameters.abc.PlayingIndicator` on any :class:`~mutwo.events.abc.Event`.

    :param playing_indicator_converters: A sequence of :class:`PlayingIndicatorConverter` which shall
        be applied on each :class:`~mutwo.events.basic.SimpleEvent`.
    :type playing_indicator_converters: typing.Sequence[PlayingIndicatorConverter]
    """

    def __init__(
        self, playing_indicator_converters: typing.Sequence[PlayingIndicatorConverter],
    ):
        self._playing_indicator_converters = playing_indicator_converters

    def _convert_simple_event(
        self,
        event_to_convert: events.basic.SimpleEvent,
        absolute_entry_delay: parameters.abc.DurationType,
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        """Convert instance of :class:`mutwo.events.basic.SimpleEvent`."""

        converted_event = [event_to_convert]

        for playing_indicator_converter in self._playing_indicator_converters:
            new_converted_event = []
            for simple_event in converted_event:
                converted_simple_event = playing_indicator_converter.convert(
                    simple_event
                )
                new_converted_event.extend(converted_simple_event)

            converted_event = new_converted_event

        return events.basic.SequentialEvent(converted_event)

    def _convert_simultaneous_event(
        self,
        simultaneous_event: events.basic.SimultaneousEvent,
        absolute_entry_delay: parameters.abc.DurationType,
    ) -> events.basic.SimultaneousEvent:
        return events.basic.SimultaneousEvent(
            super()._convert_simultaneous_event(
                simultaneous_event, absolute_entry_delay
            )
        )

    def _convert_sequential_event(
        self,
        sequential_event: events.basic.SequentialEvent,
        absolute_entry_delay: parameters.abc.DurationType,
    ) -> typing.Tuple[typing.Any, ...]:
        return events.basic.SequentialEvent(
            super()._convert_sequential_event(sequential_event, absolute_entry_delay)
        )

    def convert(self, event_to_convert: events.abc.Event) -> events.abc.Event:
        return self._convert_event(event_to_convert, 0)



"""Several constants which are used for the midi converter module.

There are basically two types of constants:
    1. Default values for :class:`mutwo.converters.frontends.MidiFileConverter`
       constructor.
    2. Values that are defined by the midi file standard.
"""

import expenvelope  # type: ignore

from mutwo import parameters

DEFAULT_AVAILABLE_MIDI_CHANNELS = tuple(range(16))
"""default value for ``available_midi_channels`` in `MidiFileConverter`"""

DEFAULT_MAXIMUM_PITCH_BEND_DEVIATION_IN_CENTS = 200
"""default value for ``maximum_pitch_bend_deviation_in_cents`` in `MidiFileConverter`"""

DEFAULT_MIDI_FILE_TYPE = 1
"""default value for ``midi_file_type`` in `MidiFileConverter`"""

DEFAULT_MIDI_INSTRUMENT_NAME = "Acoustic Grand Piano"
"""default value for ``midi_instrument_name`` in `MidiFileConverter`"""

DEFAULT_N_MIDI_CHANNELS_PER_TRACK = 1
"""default value for ``n_midi_channels_per_track`` in `MidiFileConverter`"""

DEFAULT_TEMPO_ENVELOPE: expenvelope.Envelope = expenvelope.Envelope.from_levels_and_durations(
    levels=[parameters.tempos.TempoPoint(120, 1), parameters.tempos.TempoPoint(120, 1)],
    durations=[1],
)
"""default value for ``tempo_envelope`` in `MidiFileConverter`"""

DEFAULT_TICKS_PER_BEAT = 480
"""default value for ``ticks_per_beat`` in `MidiFileConverter`"""

ALLOWED_MIDI_CHANNELS = tuple(range(16))
"""midi channels that are allowed (following the standard
midi file definition)."""

MAXIMUM_MICROSECONDS_PER_BEAT = 16777215

MIDI_TEMPO_FACTOR = 1000000
"""factor to multiply beats-in-seconds to get
beats-in-microseconds (which is the tempo unit for midi)"""

NEUTRAL_PITCH_BEND = 8191
"""the value for midi pitch bend when the resulting pitch
doesn't change"""

MAXIMUM_PITCH_BEND = 16382
"""the highest allowed value for midi pitch bend"""



"""Submodule for the parameter volume.

'Volume' is defined as any object that knows a :attr:`amplitude` attribute.
"""

import typing

import numpy as np  # type: ignore

from mutwo import parameters
from mutwo.utilities import constants
from mutwo.utilities import tools

__all__ = ("DirectVolume", "DecibelVolume", "WesternVolume")


class DirectVolume(parameters.abc.Volume):
    """A simple volume class that gets directly initialised by its amplitude.

    :param amplitude: The amplitude of the :class:`DirectVolume` object.

    May be used when a converter class needs a volume object, but there is
    no need or desire for a complex abstraction of the respective volume.
    """

    def __init__(self, amplitude: constants.Real):
        self._amplitude = amplitude

    @property
    def amplitude(self) -> constants.Real:
        return self._amplitude

    def __repr__(self) -> str:
        return "{}({})".format(type(self).__name__, self.amplitude)


class DecibelVolume(parameters.abc.Volume):
    """A simple volume class that gets directly initialised by decibel.

    :param decibel: The decibel of the :class:`DecibelVolume` object (should be
        from -120 to 0).

    May be used when a converter class needs a volume object, but there is
    no need or desire for a complex abstraction of the respective volume.
    """

    def __init__(self, decibel: constants.Real):
        self._decibel = decibel

    @property
    def decibel(self) -> constants.Real:
        return self._decibel

    @property
    def amplitude(self) -> constants.Real:
        return self.decibel_to_amplitude_ratio(self.decibel)

    def __repr__(self) -> str:
        return "{}({})".format(type(self).__name__, self.amplitude)


class WesternVolume(parameters.abc.Volume):
    """Volume with a traditional Western nomenclature.

    :param name: Dynamic indicator in traditional Western nomenclature
        ('f', 'pp', 'mf', 'sfz', etc.). For a list of all supported
        indicators, see :const:`mutwo.parameters.volumes_constants.DYNAMIC_INDICATOR`.

    >>> from mutwo.parameters import volumes
    >>> volumes.WesternVolume('fff')
    """

    def __init__(self, name: str):
        self.name = name
        self._standard_dynamic_indicator_to_decibel_mapping = (
            WesternVolume._make_standard_dynamic_indicator_to_decibel_mapping()
        )
        self._dynamic_indicator_to_decibel_mapping = WesternVolume._make_dynamic_indicator_to_decibel_mapping(
            self._standard_dynamic_indicator_to_decibel_mapping
        )
        self._decibel_to_standard_dynamic_indicator_mapping = {
            decibel: dynamic_indicator
            for dynamic_indicator, decibel in self._standard_dynamic_indicator_to_decibel_mapping.items()
        }

    def __repr__(self) -> str:
        return "{}({})".format(type(self).__name__, self.name)

    # ###################################################################### #
    #                      static private methods                            #
    # ###################################################################### #

    @staticmethod
    def _make_standard_dynamic_indicator_to_decibel_mapping() -> typing.Dict[
        str, float
    ]:
        return {
            dynamic_indicator: decibel
            for dynamic_indicator, decibel in zip(
                parameters.volumes_constants.STANDARD_DYNAMIC_INDICATOR,
                np.linspace(
                    parameters.volumes_constants.MINIMUM_DECIBEL_FOR_STANDARD_DYNAMIC_INDICATOR,
                    parameters.volumes_constants.MAXIMUM_DECIBEL_FOR_STANDARD_DYNAMIC_INDICATOR,
                    len(parameters.volumes_constants.STANDARD_DYNAMIC_INDICATOR),
                    dtype=float,
                ),
            )
        }

    @staticmethod
    def _make_dynamic_indicator_to_decibel_mapping(
        standard_dynamic_indicator_to_decibel_mapping: typing.Dict[str, float]
    ) -> typing.Dict[str, float]:
        dynamic_indicator_to_decibel_mapping = {}
        dynamic_indicator_to_decibel_mapping.update(
            standard_dynamic_indicator_to_decibel_mapping
        )
        for (
            special_dynamic_indicator,
            standard_dynamic_indicator,
        ) in (
            parameters.volumes_constants.SPECIAL_DYNAMIC_INDICATOR_TO_STANDARD_DYNAMIC_INDICATOR_MAPPING.items()
        ):
            dynamic_indicator_to_decibel_mapping.update(
                {
                    special_dynamic_indicator: dynamic_indicator_to_decibel_mapping[
                        standard_dynamic_indicator
                    ]
                }
            )
        return dynamic_indicator_to_decibel_mapping

    # ###################################################################### #
    #                class methods (alternative constructors)                #
    # ###################################################################### #

    @classmethod
    def from_amplitude(cls, amplitude: constants.Real) -> "WesternVolume":
        """Initialise `WesternVolume` from amplitude ratio.

        :param amplitude: The amplitude which shall be converted to a `WesternVolume`
            object.

        >>> from mutwo.parameters import volumes
        >>> volumes.WesternVolume.from_amplitude(0.05)
        WesternVolume(mp)
        """
        decibel = cls.amplitude_ratio_to_decibel(amplitude)
        return cls.from_decibel(decibel)

    @classmethod
    def from_decibel(cls, decibel: constants.Real) -> "WesternVolume":
        """Initialise `WesternVolume` from decibel.

        :param decibel: The decibel which shall be converted to a `WesternVolume`
            object.

        >>> from mutwo.parameters import volumes
        >>> volumes.WesternVolume.from_decibel(-24)
        WesternVolume(mf)
        """
        volume_object = cls("mf")
        closest_decibel: float = tools.find_closest_item(
            decibel,
            tuple(volume_object._decibel_to_standard_dynamic_indicator_mapping.keys()),
        )
        indicator = volume_object._decibel_to_standard_dynamic_indicator_mapping[
            closest_decibel
        ]
        volume_object.name = indicator
        return volume_object

    # ###################################################################### #
    #                             properties                                 #
    # ###################################################################### #

    @property
    def name(self) -> str:
        """The western nomenclature name for dynamic.

        For a list of all supported indicators, see
        :const:`mutwo.parameters.volumes_constants.DYNAMIC_INDICATOR`.
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        try:
            assert name in parameters.volumes_constants.DYNAMIC_INDICATOR
        except AssertionError:
            message = (
                "unknown dynamic name '{}'. Supported dynamic names are '{}'.".format(
                    name, parameters.volumes_constants.DYNAMIC_INDICATOR,
                )
            )
            raise ValueError(message)
        self._name = name

    @property
    def decibel(self) -> constants.Real:
        return self._dynamic_indicator_to_decibel_mapping[self.name]

    @property
    def amplitude(self) -> constants.Real:
        return self.decibel_to_amplitude_ratio(self.decibel)



"""Constants to be used for and with :mod:`mutwo.converters.frontends.csound`.

The file mostly contains different flags for running Csound.
The flag definitions are documented
`here <https://csound.com/docs/manual/CommandFlags.html>`_.
"""

SILENT_FLAG = "-O null"
"""Flag for preventing Csound from printing any information
while rendering."""

FORMAT_IRCAM = "--format=ircam"
"""Flag for rendering sound files in IRCAM format."""

FORMAT_24BIT = "--format=24bit"
"""Flag for rendering sound files in 24bit."""

FORMAT_64BIT = "--format=double"
"""Flag for rendering sound files in 64bit floating point."""

FORMAT_8BIT = "--format=uchar"
"""Flag for rendering sound files in 8bit."""

FORMAT_FLOAT = "--format=float"
"""Flag for rendering sound files in single-format float audio samples."""

FORMAT_WAV = "--format=wav"
"""Flag for rendering sound files in wav file format."""

SEQUENTIAL_EVENT_ANNOTATION = ";; NEW SEQUENTIAL EVENT\n;;"
"""Annotation in Csound Score files when a new :class:`SequentialEvent` starts."""

SIMULTANEOUS_EVENT_ANNOTATION = ";; NEW SIMULTANEOUS EVENT\n;;"
"""Annotation in Csound Score files when a new :class:`SimultaneousEvent` starts."""

N_EMPTY_LINES_AFTER_COMPLEX_EVENT = 1
"""How many empty lines shall be written to a Csound Score file after a :class:`ComplexEvent`."""



"""Set default values for :class:`mutwo.events.music.NoteLike`."""

from mutwo import parameters

DEFAULT_PLAYING_INDICATORS_COLLECTION_CLASS = (
    parameters.playing_indicators.PlayingIndicatorCollection
)
"""Default value for :attr:`~mutwo.events.music.NoteLike.playing_indicators`
in :class:`~mutwo.events.music.NoteLike`"""

DEFAULT_NOTATION_INDICATORS_COLLECTION_CLASS = (
    parameters.notation_indicators.NotationIndicatorCollection
)
"""Default value for :attr:`~mutwo.events.music.NoteLike.notation_indicators`
in :class:`~mutwo.events.music.NoteLike`"""



"""Definition of global variables which are used all over mutwo.

The main reason for this package is a mypy issue with Pythons buildin
[numbers module](https://docs.python.org/3/library/numbers.html) which
is documented [here](https://github.com/python/mypy/issues/3186). Mypy
doesn't accept numbers abstract base classes. Until numbers will be
supported users have to define their own typing data for general number
classes. PEP 3141 recommends users to simply annotate arguments with
'float', but this wouldn't include `fractions.Fraction` which is often
necessary in musical contexts (as github user arseniiv also remarked).
"""

import fractions
import typing

Real = typing.Union[float, fractions.Fraction]



"""Generic decorators that are used within :mod:`mutwo`."""

import copy
import functools
import typing

__all__ = ("add_return_option",)


F = typing.TypeVar("F", bound=typing.Callable[..., typing.Any])


def add_return_option(function: F) -> F:
    """This decorator adds a return option for object mutating methods.

    The 'add_return_option' decorator adds the 'mutate' keyword argument
    to the decorated method. If 'mutate' is set to False, the decorator deep
    copies the respective object, then applies the called method on the new
    copied object and finally returns the copied object. This can be useful
    for methods that by default mutate its object. When adding this method,
    it is up to the user whether the original object shall be changed
    (for mutate=True) or if a copied version of the object with the
    respective mutation shall be returned (for mutate=False).
    """

    @functools.wraps(function)
    def wrapper(self, *args, mutate: bool = True, **kwargs) -> typing.Any:
        if mutate is True:
            function(self, *args, **kwargs)
        else:
            deep_copied_object = copy.deepcopy(self)
            function(deep_copied_object, *args, **kwargs)
            return deep_copied_object

    wrapped_function = typing.cast(F, wrapper)
    wrapped_function.__annotations__.update({"mutate": bool})
    return wrapped_function



"""Generic event classes which can be used in multiple contexts.

The different events differ in their timing structure and whether they
are nested or not:
"""

from __future__ import annotations

import bisect
import copy
import types
import typing

from mutwo import events
from mutwo import parameters
from mutwo.utilities import constants
from mutwo.utilities import decorators
from mutwo.utilities import tools


__all__ = ("SimpleEvent", "SequentialEvent", "SimultaneousEvent")


class SimpleEvent(events.abc.Event):
    """Event-Object which doesn't contain other Event-Objects (the node or leaf).

    :param new_duration: The duration of the ``SimpleEvent``. This can be any number.
        The unit of the duration is up to the interpretation of the user and the
        respective converter routine that will be used.

    **Example:**

    >>> from mutwo.events import basic
    >>> simple_event = basic.SimpleEvent(2)
    >>> print(simple_event)
    SimpleEvent(duration = 2)
    """

    def __init__(self, new_duration: parameters.abc.DurationType):
        self.duration = new_duration

    # ###################################################################### #
    #                           magic methods                                #
    # ###################################################################### #

    def __eq__(self, other: typing.Any) -> bool:
        """Test for checking if two objects are equal."""
        try:
            return self._is_equal(other) and other._is_equal(self)
        except AttributeError:
            return False

    def __repr__(self) -> str:
        attributes = (
            "{} = {}".format(attribute, getattr(self, attribute))
            for attribute in self._parameters_to_print
        )
        return "{}({})".format(type(self).__name__, ", ".join(attributes))

    # ###################################################################### #
    #                           properties                                   #
    # ###################################################################### #

    @property
    def _parameters_to_print(self) -> typing.Tuple[str, ...]:
        """Return tuple of attribute names which shall be printed for repr.
        """
        return self._parameters_to_compare

    @property
    def _parameters_to_compare(self) -> typing.Tuple[str, ...]:
        """Return tuple of attribute names which values define the SimpleEvent.

        The returned attribute names are used for equality check between two
        SimpleEvent objects.
        """
        return tuple(
            attribute
            for attribute in dir(self)
            # no private attributes
            if attribute[0] != "_"
            # no methods
            and not isinstance(getattr(self, attribute), types.MethodType)
        )

    @property
    def duration(self) -> parameters.abc.DurationType:
        return self._duration

    @duration.setter
    def duration(self, new_duration: parameters.abc.DurationType):
        self._duration = new_duration

    # ###################################################################### #
    #                           private methods                              #
    # ###################################################################### #

    def _is_equal(self, other: typing.Any) -> bool:
        """Helper function to inspect if two SimpleEvent objects are equal."""

        for parameter_to_compare in self._parameters_to_compare:
            try:
                # if the assigned values of the specific parameter aren't
                # equal, both objects can't be equal
                if getattr(self, parameter_to_compare) != getattr(
                    other, parameter_to_compare
                ):
                    return False

            # if the other object doesn't know the essential parameter
            # mutwo assumes that both objects can't be equal
            except AttributeError:
                return False

        # if all compared parameters are equal, return True
        return True

    # ###################################################################### #
    #                           public methods                               #
    # ###################################################################### #

    def destructive_copy(self) -> SimpleEvent:
        return copy.deepcopy(self)

    def get_parameter(self, parameter_name: str) -> typing.Any:
        """Return event attribute with the entered name.

        :parameter_name: The name of the attribute that shall be returned.
        :returns: Return the value that has been assigned to the passed
            parameter_name. If an event doesn't posses the asked parameter,
            the method will simply return None.

        **Example:**

        >>> from mutwo.events import basic
        >>> simple_event = basic.SimpleEvent(2)
        >>> simple_event.pitch = 200
        >>> simple_event.get_parameter('pitch')
        200
        >>> simple_event.get_parameter('pitches')
        None
        """
        try:
            return getattr(self, parameter_name)
        except AttributeError:
            return None

    @decorators.add_return_option
    def set_parameter(  # type: ignore
        self,
        parameter_name: str,
        object_or_function: typing.Union[
            typing.Callable[[parameters.abc.Parameter], parameters.abc.Parameter],
            typing.Any,
        ],
        set_unassigned_parameter: bool = True,
    ) -> typing.Optional[SimpleEvent]:
        """Sets event parameter to new value.

        :param parameter_name: The name of the parameter which values shall be changed.
        :param object_or_function: For setting the parameter either a new value can be
            passed directly or a function can be passed. The function gets as an
            argument the previous value that has had been assigned to the respective
            object and has to return a new value that will be assigned to the object.
        :param set_unassigned_parameter: If set to False a new parameter will only be
            assigned to an Event if the Event already has a attribute with the
            respective `parameter_name`. If the Event doesn't know the attribute yet
            and `set_unassigned_parameter` is False, the method call will simply be
            ignored.

        **Example:**

        >>> from mutwo.events import basic
        >>> simple_event = basic.SimpleEvent(2)
        >>> simple_event.set_parameter('duration', lambda old_duration: old_duration * 2)
        >>> simple_event.duration
        4
        >>> simple_event.set_parameter('duration', 3)
        >>> simple_event.duration
        3
        >>> simple_event.set_parameter('unknown_parameter', 10, set_unassigned_parameter=False)  # this will be ignored
        >>> simple_event.unknown_parameter
        AttributeError: 'SimpleEvent' object has no attribute 'unknown_parameter'
        >>> simple_event.set_parameter('unknown_parameter', 10, set_unassigned_parameter=True)  # this will be written
        >>> simple_event.unknown_parameter
        10
        """

        old_parameter = self.get_parameter(parameter_name)
        if set_unassigned_parameter or old_parameter is not None:
            if hasattr(object_or_function, "__call__"):
                new_parameter = object_or_function(old_parameter)
            else:
                new_parameter = object_or_function
            setattr(self, parameter_name, new_parameter)

    @decorators.add_return_option
    def mutate_parameter(  # type: ignore
        self,
        parameter_name: str,
        function: typing.Union[
            typing.Callable[[parameters.abc.Parameter], None], typing.Any
        ],
    ) -> typing.Optional[SimpleEvent]:
        parameter = self.get_parameter(parameter_name)
        if parameter is not None:
            function(parameter)

    @decorators.add_return_option
    def cut_out(  # type: ignore
        self, start: parameters.abc.DurationType, end: parameters.abc.DurationType,
    ) -> typing.Optional[SimpleEvent]:
        self._assert_correct_start_and_end_values(
            start, end, condition=lambda start, end: start < end
        )

        duration = self.duration

        difference_to_duration: constants.Real = 0

        if start > 0:
            difference_to_duration += start
        if end < duration:
            difference_to_duration += duration - end

        try:
            assert difference_to_duration < duration
        except AssertionError:
            message = (
                "Can't cut out SimpleEvent '{}' with duration '{}' from (start = {} to"
                " end = {}).".format(self, duration, start, end)
            )
            raise ValueError(message)

        self.duration -= difference_to_duration

    @decorators.add_return_option
    def cut_off(  # type: ignore
        self, start: parameters.abc.DurationType, end: parameters.abc.DurationType,
    ) -> typing.Optional[SimpleEvent]:
        self._assert_correct_start_and_end_values(start, end)

        duration = self.duration
        if start < duration:
            if end > duration:
                end = duration

            self.duration -= end - start


T = typing.TypeVar("T", bound=events.abc.Event)


class SequentialEvent(events.abc.ComplexEvent, typing.Generic[T]):
    """Event-Object which contains other Events which happen in a linear order."""

    # ###################################################################### #
    #                           properties                                   #
    # ###################################################################### #

    @events.abc.ComplexEvent.duration.getter
    def duration(self) -> parameters.abc.DurationType:
        return sum(event.duration for event in self)

    @property
    def absolute_times(self) -> typing.Tuple[constants.Real, ...]:
        """Return absolute point in time for each event."""

        durations = (event.duration for event in self)
        return tuple(tools.accumulate_from_zero(durations))[:-1]

    @property
    def start_and_end_time_per_event(
        self,
    ) -> typing.Tuple[typing.Tuple[constants.Real, constants.Real], ...]:
        """Return start and end time for each event."""

        durations = (event.duration for event in self)
        absolute_times = tuple(tools.accumulate_from_zero(durations))
        return tuple(zip(absolute_times, absolute_times[1:]))

    # ###################################################################### #
    #                           public methods                               #
    # ###################################################################### #

    def get_event_index_at(self, absolute_time: constants.Real) -> int:
        """Get index of event which is active at the passed absolute_time.

        :param absolute_time: The absolute time where the method shall search
            for the active event.

        **Example:**

        >>> from mutwo.events import basic
        >>> sequential_event = basic.SequentialEvent([basic.SimpleEvent(2), basic.SimpleEvent(3)])
        >>> sequential_event.get_event_index_at(1)
        0
        >>> sequential_event.get_event_index_at(3)
        1
        """

        absolute_times = self.absolute_times
        return bisect.bisect_right(absolute_times, absolute_time) - 1

    def get_event_at(self, absolute_time: constants.Real) -> T:
        """Get event which is active at the passed absolute_time.

        :param absolute_time: The absolute time where the method shall search
            for the active event.

        **Example:**

        >>> from mutwo.events import basic
        >>> sequential_event = basic.SequentialEvent([basic.SimpleEvent(2), basic.SimpleEvent(3)])
        >>> sequential_event.get_event_at(1)
        SimpleEvent(duration = 2)
        >>> sequential_event.get_event_at(3)
        SimpleEvent(duration = 3)
        """

        return self[self.get_event_index_at(absolute_time)]  # type: ignore

    @decorators.add_return_option
    def cut_out(  # type: ignore
        self, start: parameters.abc.DurationType, end: parameters.abc.DurationType,
    ) -> typing.Optional[SequentialEvent[T]]:
        self._assert_correct_start_and_end_values(start, end)

        remove_nth_event = []
        for nth_event, event_start, event in zip(
            range(len(self)), self.absolute_times, self
        ):
            event_duration = event.duration
            event_end = event_start + event_duration

            cut_out_start: constants.Real = 0
            cut_out_end = event_duration

            if event_start < start:
                cut_out_start += start - event_start

            if event_end > end:
                cut_out_end -= event_end - end

            if cut_out_start < cut_out_end:
                event.cut_out(cut_out_start, cut_out_end)
            else:
                remove_nth_event.append(nth_event)

        for nth_event_to_remove in reversed(remove_nth_event):
            del self[nth_event_to_remove]

    @decorators.add_return_option
    def cut_off(  # type: ignore
        self, start: parameters.abc.DurationType, end: parameters.abc.DurationType,
    ) -> typing.Optional[SequentialEvent[T]]:
        cut_off_duration = end - start

        # avoid unnecessary iterations
        if cut_off_duration > 0:

            # collect events which are only active within the
            # cut_off - range
            events_to_delete = []
            for event_index, event_start, event in zip(
                range(len(self)), self.absolute_times, self
            ):
                event_end = event_start + event.duration
                if event_start >= start and event_end <= end:
                    events_to_delete.append(event_index)

                # shorten event which are partly active within the
                # cut_off - range
                elif event_start <= start and event_end >= start:
                    difference_to_event_start = start - event_start
                    event.cut_off(
                        difference_to_event_start,
                        difference_to_event_start + cut_off_duration,
                    )

                elif event_start < end and event_end > end:
                    difference_to_event_start = event_start - start
                    event.cut_off(0, cut_off_duration - difference_to_event_start)

            for index in reversed(events_to_delete):
                del self[index]

    @decorators.add_return_option
    def squash_in(  # type: ignore
        self, start: parameters.abc.DurationType, event_to_squash_in: events.abc.Event
    ) -> typing.Optional[SequentialEvent[T]]:
        self._assert_start_in_range(start)

        cut_off_end = start + event_to_squash_in.duration

        self.cut_off(start, cut_off_end)

        if start == self.duration:
            self.append(event_to_squash_in)

        else:
            absolute_times = self.absolute_times
            active_event_index = self.get_event_index_at(start)
            split_position = start - absolute_times[active_event_index]
            # potentially split event
            if split_position > 6e-14:  # avoid floating point errors
                split_active_event = self[active_event_index].split_at(split_position)
                self[active_event_index] = split_active_event[1]
                self.insert(active_event_index, split_active_event[0])
                active_event_index += 1

            self.insert(active_event_index, event_to_squash_in)


class SimultaneousEvent(events.abc.ComplexEvent, typing.Generic[T]):
    """Event-Object which contains other Event-Objects which happen at the same time."""

    # ###################################################################### #
    #                           properties                                   #
    # ###################################################################### #

    @events.abc.ComplexEvent.duration.getter
    def duration(self) -> parameters.abc.DurationType:
        return max(event.duration for event in self)

    # ###################################################################### #
    #                           public methods                               #
    # ###################################################################### #

    @decorators.add_return_option
    def cut_out(  # type: ignore
        self, start: parameters.abc.DurationType, end: parameters.abc.DurationType,
    ) -> typing.Optional[SimultaneousEvent[T]]:
        self._assert_correct_start_and_end_values(start, end)
        [event.cut_out(start, end) for event in self]

    @decorators.add_return_option
    def cut_off(  # type: ignore
        self, start: parameters.abc.DurationType, end: parameters.abc.DurationType,
    ) -> typing.Optional[SimultaneousEvent[T]]:
        self._assert_correct_start_and_end_values(start, end)
        [event.cut_off(start, end) for event in self]

    @decorators.add_return_option
    def squash_in(  # type: ignore
        self, start: parameters.abc.DurationType, event_to_squash_in: events.abc.Event
    ) -> typing.Optional[SimultaneousEvent[T]]:
        self._assert_start_in_range(start)

        for event in self:
            try:
                event.squash_in(start, event_to_squash_in)  # type: ignore

            # simple events don't have a 'squash_in' method
            except AttributeError:
                message = (
                    "Can't squash '{}' in '{}'. Does the SimultaneousEvent contain"
                    " events that inherit from SimpleEvent? For being able to"
                    " squash in, the SimultaneousEvent needs to contain"
                    " SequentialEvents or SimultaneousEvents.".format(
                        event_to_squash_in, self
                    )
                )
                raise TypeError(message)



"""Several subpackages to convert data from and to mutwo.

The subpackages differentiate in their conversion mapping:
    - backends:     external data                   **->** mutwo or generic Python objects
    - frontends:    mutwo or generic Python objects **->** external data
    - symmetrical:  mutwo or generic Python objects **->** mutwo or generic Python objects

*External data* can be a file, a Python object from a third-party library or a Stream
object (that can be started to play for triggering events in real-time).
"""

from . import abc
from . import frontends
from . import backends
from . import symmetrical



"""Algorithms which are related to British composer Michael Edwards."""

import functools
import itertools
import operator

from mutwo.generators import edwards_constants
from mutwo.utilities import tools

__all__ = ("ActivityLevel",)


class ActivityLevel(object):
    """Python implementation of Michael Edwards activity level algorithm.

    :param start_at: from which pattern per level shall be started (can be
        either 0, 1 or 2)

    Activity Levels is a concept derived from Michael Edwards.
    Quoting Michael Edwards, Activity Levels are an `"object for determining
    (deterministically) on a call-by-call basis whether a process is active
    or not (boolean).  This is determined by nine 10-element lists
    (actually three versions of each) of hand-coded 1s and 0s, each list
    representing an 'activity-level' (how active the process should be).
    The first three 10-element lists have only one 1 in them, the rest being zeros.
    The second three have two 1s, etc. Activity-levels of 0 and 10 would return
    never active and always active respectively."
    <https://michael-edwards.org/sc/robodoc/activity-levels_lsp.html#robo23>`_.

    **Example:**

    >>> from mutwo.generators import edwards
    >>> activity_levels = edwards.ActivityLevel()
    >>> activity_levels(0)  # activity level 0 will always return False
    False
    >>> activity_levels(10)  # activity level 10 will always return True
    True
    >>> activity_levels(7)  # activity level 7 will mostly return True
    True
    >>> tuple(activity_levels(7) for _ in range(10))
    (True, False, True, True, False, True, True, False, True, True)
    """

    _allowed_range = tuple(range(len(edwards_constants.ACTIVITY_LEVELS)))

    def __init__(self, start_at: int = 0) -> None:
        try:
            assert start_at in (0, 1, 2)
        except AssertionError:
            msg = "start_at has to be either 0, 1 or 2 and not {}, ".format(start_at)
            msg += "because there are only three different tuples defined per level."
            raise ValueError(msg)

        self._activity_level_cycles = tuple(
            itertools.cycle(
                functools.reduce(
                    operator.add, tuple(tools.cyclic_permutations(levels))[start_at]
                )
            )
            for levels in edwards_constants.ACTIVITY_LEVELS
        )

    def __repr__(self) -> str:
        return "ActivityLevel()"

    def __call__(self, level: int) -> bool:
        """Return current state (is active or not) of entered activity level.

        :param level: the activity-level which current state shall be returned
            (should be from 0 to 10)
        :return: True if active and False if not active.
        """

        try:
            assert level in self._allowed_range
        except AssertionError:
            msg = "level is '{}' but has to be in range '{}'!".format(
                level, self._allowed_range
            )
            raise ValueError(msg)

        return bool(next(self._activity_level_cycles[level]))



"""This module adds several constants that are used for the different volume classes.

"""

MINIMUM_VELOCITY = 0
"""the lowest allowed midi velocity value"""

MAXIMUM_VELOCITY = 127
"""the highest allowed midi velocity value"""

# standard volume indicator
STANDARD_DYNAMIC_INDICATOR = tuple(
    "ppppp pppp ppp pp p mp mf f ff fff ffff fffff".split(" ")
)

MINIMUM_DECIBEL_FOR_STANDARD_DYNAMIC_INDICATOR: float = -52
"""decibel value for lowest dynamic indicator ('ppppp')"""

MAXIMUM_DECIBEL_FOR_STANDARD_DYNAMIC_INDICATOR: float = 0
"""decibel value for highest dynamic indicator ('fffff')"""

SPECIAL_DYNAMIC_INDICATOR_TO_STANDARD_DYNAMIC_INDICATOR_MAPPING = {
    "fp": "mf",
    "sf": "f",
    "sff": "ff",
    "sfz": "ff",
    "sp": "p",
    "spp": "pp",
    "rfz": "f",
}

DYNAMIC_INDICATOR = STANDARD_DYNAMIC_INDICATOR + tuple(
    SPECIAL_DYNAMIC_INDICATOR_TO_STANDARD_DYNAMIC_INDICATOR_MAPPING.keys()
)
"""all available dynamic indicator for :class:`WesternVolume`"""



"""Constants to be used for and with :mod:`mutwo.converters.frontends.abjad`.
"""

import inspect

from mutwo.converters.frontends import abjad_attachments


DEFAULT_ABJAD_ATTACHMENT_CLASSES = tuple(
    cls
    for _, cls in inspect.getmembers(abjad_attachments, inspect.isclass)
    if not inspect.isabstract(cls)
)
"""Default value for argument `abjad_attachment_classes` in
:class:`~mutwo.converters.frontends.SequentialEventToAbjadVoiceConverter`."""
